#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include "lua.h"
#include "lauxlib.h"

#define MYNAME "inotify"

enum
{
	STACKBUF = 4096
};

typedef struct Inotify Inotify;
typedef struct Inotifyflag Inotifyflag;

struct Inotify
{
	int fd;
};

struct Inotifyflag
{
	const char *name;
	uint32_t flag;
	int isbit;
};

static Inotifyflag linotify_flags[] =
{
#undef flag
#define flag(f) { #f, f, 1 }
	flag(IN_ACCESS),
	flag(IN_ATTRIB),
	flag(IN_CLOSE_WRITE),
	flag(IN_CLOSE_NOWRITE),
	flag(IN_CREATE),
	flag(IN_DELETE),
	flag(IN_DELETE_SELF),
	flag(IN_MODIFY),
	flag(IN_MOVE_SELF),
	flag(IN_MOVED_FROM),
	flag(IN_MOVED_TO),
	flag(IN_OPEN),
	flag(IN_IGNORED),
	flag(IN_ISDIR),
	flag(IN_Q_OVERFLOW),
	flag(IN_UNMOUNT),
#undef flag
#define flag(f) { #f, f, 0 }
	flag(IN_DONT_FOLLOW),
	flag(IN_MASK_ADD),
	flag(IN_ONESHOT),
	flag(IN_ONLYDIR),
	flag(IN_CLOSE),
	flag(IN_MOVE),
	flag(IN_ALL_EVENTS),
	{ 0, 0, 0 }
#undef flag
};

static void
xfree(void *ptr)
{
	if(ptr)
		free(ptr);
}

static Inotify *
linotify_get(lua_State *L, int idx)
{
	Inotify *inotify;

	inotify = lua_touserdata(L, idx);
	if(inotify && lua_getmetatable(L, idx) && lua_rawequal(L, -1, LUA_ENVIRONINDEX)){
		lua_pop(L, 1);
		return inotify;
	}
	luaL_typerror(L, idx, MYNAME);
	return 0;
}

static int
linotify_getfd(lua_State *L, int idx)
{
	if(lua_isnumber(L, idx))
		return lua_tointeger(L, idx);
	else
		return linotify_get(L, idx)->fd;
}

static int
linotify_init(lua_State *L)
{
	int fd;
	Inotify *inotify;

	fd = inotify_init();
	if(fd < 0){
		int xerrno = errno;
		lua_pushnil(L);
		lua_pushnumber(L, xerrno);
		lua_pushstring(L, strerror(xerrno));
		return 3;
	}
	if(lua_toboolean(L, 1)){
		inotify = lua_newuserdata(L, sizeof(*inotify));
		inotify->fd = fd;
		lua_pushvalue(L, LUA_ENVIRONINDEX);
		lua_setmetatable(L, -2);
	}else
		lua_pushnumber(L, fd);
	return 1;
}

static int
linotify_close(lua_State *L)
{
	int fd, ret, nret=0;
	Inotify *inotify=0;

	if(lua_isnumber(L, 1))
		fd = (int)lua_tointeger(L, 1);
	else{
		inotify = linotify_get(L, 1);
		fd = inotify->fd;
	}
	ret = close(fd);
	if(ret < 0){
		int xerrno = errno;
		lua_pushnumber(L, xerrno);
		lua_pushstring(L, strerror(xerrno));
		nret = 2;
	}
	if(inotify){
		lua_pushnil(L);
		lua_setmetatable(L, 1);
	}
	return nret;
}

static uint32_t
linotify_getflags(lua_State *L, int narg, const char *name, uint32_t def, int isopt, Inotifyflag flags[])
{
	int i, j, len;
	uint32_t flagz=0;
	const char *flag;

	if(lua_isnoneornil(L, narg) && isopt)
		return def;
	if(lua_type(L, narg) != LUA_TTABLE)
		luaL_error(L, "bad argument '%s' to inotify.add_watch: table expected, got %s", name, lua_typename(L, lua_type(L, narg)));
	len = lua_objlen(L, narg);
	for(i=1; i<=len; i++){
		lua_rawgeti(L, narg, i);
		flag = lua_tostring(L, -1);
		for(j=0; flags[j].name; j++){
			if(!strcmp(flag, flags[j].name)){
				flagz |= flags[j].flag;
				break;
			}
		}
		if(!flags[j].name)
			luaL_error(L, "bad argument '%s' to inotify.add_watch: unknown flag '%s'", name, flag);
		lua_pop(L, 1);
	}

	return flagz;
}

static int
linotify_add_watch(lua_State *L)
{
	int fd, wd;
	uint32_t mask;
	const char *pathname;

	fd = linotify_getfd(L, 1);
	pathname = luaL_checklstring(L, 2, 0);
	mask = linotify_getflags(L, 3, "mask", IN_ALL_EVENTS, 1, linotify_flags);
	wd = inotify_add_watch(fd, pathname, mask);
	if(wd < 0){
		int xerrno = errno;
		lua_pushnil(L);
		lua_pushnumber(L, xerrno);
		lua_pushstring(L, strerror(xerrno));
		return 3;
	}
	lua_pushnumber(L, wd);
	return 1;
}

static int
linotify_rm_watch(lua_State *L)
{
	int fd, wd, ret;

	fd = linotify_getfd(L, 1);
	wd = luaL_checkint(L, 2);
	ret = inotify_rm_watch(fd, wd);
	if(ret < 0){
		int xerrno = errno;
		lua_pushnumber(L, xerrno);
		lua_pushstring(L, strerror(xerrno));
		return 2;
	}
	return 0;
}

static int
linotify_xread(lua_State *L, int nonblock)
{
	int fd, nread, nreadmax, xerrno, eventidx;
	char stackbuf[STACKBUF], *buf=stackbuf, *bufptr;

	fd = linotify_getfd(L, 1);
	nreadmax = luaL_optint(L, 2, 0);
	if(ioctl(fd, FIONREAD, &nread) < 0)
		goto error;
	if((nread == 0) && nonblock)
		return 0;
	if(nread == 0)
		nread = STACKBUF;
	if((nreadmax > 0) && (nread > nreadmax))
		nread = nreadmax;
	if((nread > STACKBUF) && !(buf = malloc(nread)))
		goto error;
	nread = read(fd, buf, nread);
	if(nread == 0)		/* kernel<2.6.21 (see inotify(7)) */
		errno = EINVAL;
	if(nread <= 0)
		goto error;
	lua_newtable(L);
	for(bufptr=buf, eventidx=0; bufptr<buf+nread;){
		struct inotify_event *in = (struct inotify_event *)bufptr;
		lua_newtable(L);
		lua_pushliteral(L, "wd");
		lua_pushnumber(L, in->wd);
		lua_rawset(L, -3);
		lua_pushliteral(L, "mask");
		lua_pushnumber(L, in->mask);
		lua_rawset(L, -3);
		lua_pushliteral(L, "cookie");
		lua_pushnumber(L, in->cookie);
		lua_rawset(L, -3);
		lua_pushliteral(L, "name");
		lua_pushlstring(L, in->name, in->len);
		lua_rawset(L, -3);
		lua_rawseti(L, -2, ++eventidx);
		bufptr += sizeof(*in) + in->len;
	}
	if(buf != stackbuf)
		xfree(buf);
	lua_pushnumber(L, nread);
	return 2;

error:	xerrno = errno;
	if(buf != stackbuf)
		xfree(buf);
	lua_pushnil(L);
	lua_pushnil(L);
	lua_pushnumber(L, xerrno);
	lua_pushstring(L, strerror(xerrno));
	return 4;
}

static int linotify_read(lua_State *L)		{ return linotify_xread(L, 0); }
static int linotify_nbread(lua_State *L)	{ return linotify_xread(L, 1); }

static int
linotify_getmaskevents(lua_State *L)
{
	uint32_t mask;
	int i, maskidx;

	mask = luaL_checkinteger(L, 1);
	lua_newtable(L);
	for(i=0, maskidx=0; linotify_flags[i].name; i++){
		if(linotify_flags[i].isbit && (mask & linotify_flags[i].flag)){
			lua_pushstring(L, linotify_flags[i].name);
			lua_rawseti(L, -2, ++maskidx);
		}
	}
	return 1;
}

static int
linotify_fd(lua_State *L)
{
	lua_pushnumber(L, linotify_get(L, 1)->fd);
	return 1;
}

static int
linotify_nread(lua_State *L)
{
	int fd, nread;

	fd = linotify_getfd(L, 1);
	if(ioctl(fd, FIONREAD, &nread) < 0){
		int xerrno = errno;
		lua_pushnil(L);
		lua_pushnumber(L, xerrno);
		lua_pushstring(L, strerror(xerrno));
		return 3;
	}
	lua_pushnumber(L, nread);
	return 1;
}

static int
linotify_tostring(lua_State *L)
{
	char s[64];

	lua_pushlstring(L, s, snprintf(s, sizeof(s), MYNAME " %p", (void *)linotify_get(L, 1)));
	return 1;
}

static struct luaL_reg linotify_methods[] =
{
	{ "__gc",		linotify_close },
	{ "__tostring",		linotify_tostring },
	{ "close",		linotify_close },
	{ "add_watch",		linotify_add_watch },
	{ "rm_watch",		linotify_rm_watch },
	{ "read",		linotify_read },
	{ "nbread",		linotify_nbread },
	{ "fd",			linotify_fd },
	{ "nread",		linotify_nread },
	{ 0, 0 }
};

static struct luaL_reg linotify_funcs[] =
{
	{ "init",		linotify_init },
	{ "close",		linotify_close },
	{ "add_watch",		linotify_add_watch },
	{ "rm_watch",		linotify_rm_watch },
	{ "read",		linotify_read },
	{ "nbread",		linotify_nbread },
	{ "getmaskevents",	linotify_getmaskevents },
	{ "nread",		linotify_nread },
	{ 0, 0 }
};

int
luaopen_inotify(lua_State *L)
{
	lua_newtable(L);
	lua_pushvalue(L, -1);
	lua_replace(L, LUA_ENVIRONINDEX);
	lua_pushvalue(L, -1);
	lua_setfield(L, -2, "__index");
	luaL_register(L, 0, linotify_methods);
	luaL_register(L, MYNAME, linotify_funcs);
	return 1;
}
